import { RPC, RequestOf, ResponseOf, RPCs } from '@previewjs/api';
import { ValueType, TypeAnalyzer, CollectedTypes } from '@previewjs/type-analyzer';
import { Reader } from '@previewjs/vfs';
import { Logger } from 'pino';
import { SerializableValue } from '@previewjs/serializable-values';
import ts from 'typescript';
import vite from 'vite';
import * as express from 'express';

type PackageDependencies = Record<string, {
    /** @deprecated */
    majorVersion: number;
    readInstalledVersion(): Promise<string | null>;
}>;

interface FrameworkPluginFactory {
    isCompatible(dependencies: PackageDependencies): Promise<boolean>;
    create(options: {
        rootDirPath: string;
        logger: Logger;
        dependencies: PackageDependencies;
    }): Promise<FrameworkPlugin>;
}
interface FrameworkPlugin {
    readonly pluginApiVersion?: number;
    readonly name: string;
    readonly defaultWrapperPath: string;
    readonly previewDirPath: string;
    readonly transformReader?: (reader: Reader) => Reader;
    readonly tsCompilerOptions?: Partial<ts.CompilerOptions>;
    readonly specialTypes?: Record<string, ValueType>;
    readonly viteConfig: (configuredPlugins: vite.Plugin[]) => vite.UserConfig;
    readonly detectComponents: (reader: Reader, typeAnalyzer: TypeAnalyzer, absoluteFilePaths: string[]) => Promise<AnalyzableComponent[]>;
}
interface AnalyzableComponent {
    readonly componentId: string;
    readonly offsets: Array<[start: number, end: number]>;
    readonly info: ComponentTypeInfo;
}
type ComponentTypeInfo = {
    kind: "component";
    readonly exported: boolean;
    readonly analyze: () => Promise<ComponentAnalysis>;
} | {
    kind: "story";
    readonly args: {
        start: number;
        end: number;
        value: SerializableValue;
    } | null;
    readonly associatedComponent: StoryAssociatedComponent;
};
type StoryAssociatedComponent = {
    readonly componentId: string;
    readonly analyze: () => Promise<ComponentAnalysis>;
};
interface ComponentAnalysis {
    propsType: ValueType;
    types: CollectedTypes;
}

type RegisterRPC = <E extends RPC<any, any>>(endpoint: E, handler: RequestHandler<RequestOf<E>, ResponseOf<E>>) => void;
type RequestHandler<Req, Res> = (request: Req) => Promise<Res>;

type SetupPreviewEnvironment = (options: {
    versionCode?: string;
    registerRPC: RegisterRPC;
    workspace: Workspace;
}) => Promise<{
    middlewares?: express.RequestHandler[];
}>;

declare function setupFrameworkPlugin({ rootDirPath, frameworkPlugins, logger, }: {
    rootDirPath: string;
    frameworkPlugins: FrameworkPluginFactory[];
    logger: Logger;
}): Promise<FrameworkPlugin | null>;

declare function createWorkspace({ rootDirPath, reader, frameworkPlugin, logger, setupEnvironment, }: {
    rootDirPath: string;
    frameworkPlugin: FrameworkPlugin;
    logger: Logger;
    reader: Reader;
    setupEnvironment?: SetupPreviewEnvironment;
}): Promise<Workspace>;
/**
 * Returns the absolute directory path of the closest ancestor containing package.json.
 */
declare function findWorkspaceRoot(absoluteFilePath: string): string | null;
interface Workspace {
    rootDirPath: string;
    reader: Reader;
    typeAnalyzer: TypeAnalyzer;
    detectComponents(options?: RequestOf<typeof RPCs.DetectComponents>): Promise<ResponseOf<typeof RPCs.DetectComponents>>;
    computeProps(options: RequestOf<typeof RPCs.ComputeProps>): Promise<ResponseOf<typeof RPCs.ComputeProps>>;
    preview: {
        start(allocatePort?: () => Promise<number>): Promise<Preview>;
    };
    dispose(): Promise<void>;
}
interface Preview {
    url(): string;
    stop(options?: {
        onceUnused?: boolean;
    }): Promise<void>;
}

export { AnalyzableComponent, ComponentAnalysis, ComponentTypeInfo, FrameworkPlugin, FrameworkPluginFactory, PackageDependencies, Preview, SetupPreviewEnvironment, Workspace, createWorkspace, findWorkspaceRoot, setupFrameworkPlugin };
