import { decodeComponentId, RPCs } from '@previewjs/api';
import { createTypeAnalyzer } from '@previewjs/type-analyzer';
import express from 'express';
import fs$1 from 'fs-extra';
import getPort, { portNumbers } from 'get-port';
import { createRequire } from 'module';
import path from 'path';
import { exclusivePromiseRunner } from 'exclusive-promises';
import fs from 'fs';
import url from 'url';
import { globby } from 'globby';
import { PREVIEW_CONFIG_NAME, readConfig } from '@previewjs/config';
import { createStackedReader, createFileSystemReader } from '@previewjs/vfs';
import assertNever from 'assert-never';
import axios from 'axios';
import { escape } from 'html-escaper';
import { createHttpTerminator } from 'http-terminator';
import viteTsconfigPaths from '@fwouts/vite-tsconfig-paths';
import fakeExportedTypesPlugin from 'rollup-plugin-friendly-type-imports';
import { loadTsconfig } from 'tsconfig-paths/lib/tsconfig-loader.js';
import * as vite from 'vite';
import { transformWithEsbuild, searchForWorkspaceRoot } from 'vite';
import { Parser } from 'acorn';
import jsx from 'acorn-jsx';

function getCacheDir(rootDirPath) {
  try {
    const __dirname = url.fileURLToPath(new URL(".", import.meta.url));
    const { version } = JSON.parse(
      fs.readFileSync(path.resolve(__dirname, "..", "package.json"), "utf8")
    );
    return path.resolve(
      rootDirPath,
      "node_modules",
      ".previewjs",
      `v${version}`
    );
  } catch (e) {
    throw new Error(`Unable to detect @previewjs/core version.
${e}`);
  }
}

async function findFiles(rootDirPath, pattern, options = {}) {
  const gitRootPath = await findGitRoot(rootDirPath);
  const relativePath = path.relative(gitRootPath, rootDirPath);
  const relativePrefix = relativePath ? relativePath + path.sep : "";
  const files = await globby(relativePrefix + pattern, {
    gitignore: true,
    ignore: ["**/node_modules/**"],
    cwd: gitRootPath,
    absolute: true,
    followSymbolicLinks: false,
    suppressErrors: true,
    deep: options.maxDepth
  });
  let normalizedRootDirPath = rootDirPath.replace(/\\/g, "/");
  if (!normalizedRootDirPath.endsWith("/")) {
    normalizedRootDirPath += "/";
  }
  return files.filter((f) => f.startsWith(normalizedRootDirPath));
}
async function findGitRoot(dirPath, fallback = dirPath) {
  try {
    if (await fs$1.exists(path.join(dirPath, ".git"))) {
      return dirPath;
    } else {
      const parentDirPath = path.dirname(dirPath);
      if (!parentDirPath || parentDirPath === dirPath) {
        return fallback;
      }
      return findGitRoot(parentDirPath, fallback);
    }
  } catch {
    return fallback;
  }
}

const FILES_REQUIRING_REDETECTION = /* @__PURE__ */ new Set([
  "jsconfig.json",
  "tsconfig.json",
  "package.json",
  "package-lock.json",
  "pnpm-lock.yaml",
  "yarn.lock"
]);
const oneAtATime = exclusivePromiseRunner();
function detectComponents(logger, workspace, frameworkPlugin, typeAnalyzer, options = {}) {
  return oneAtATime(async () => {
    logger.debug(
      `Detecting components with options: ${JSON.stringify(options)}`
    );
    const detectionStartTimestamp = Date.now();
    const cacheFilePath = path.join(
      getCacheDir(workspace.rootDirPath),
      "components.json"
    );
    const absoluteFilePaths = await (async () => {
      if (options.filePaths) {
        return options.filePaths.map(
          (filePath) => path.join(workspace.rootDirPath, filePath)
        );
      } else {
        logger.debug(
          `Finding component files from root: ${workspace.rootDirPath}`
        );
        const filePaths = await findFiles(
          workspace.rootDirPath,
          "**/*.@(js|jsx|ts|tsx|svelte|vue)"
        );
        logger.debug(`Found ${filePaths.length} component files`);
        return filePaths;
      }
    })();
    const filePathsSet = new Set(
      absoluteFilePaths.map(
        (absoluteFilePath) => path.relative(workspace.rootDirPath, absoluteFilePath).replace(/\\/g, "/")
      )
    );
    let existingCache = {
      detectionStartTimestamp: 0,
      components: []
    };
    if (fs$1.existsSync(cacheFilePath)) {
      try {
        existingCache = JSON.parse(
          fs$1.readFileSync(cacheFilePath, "utf8")
        );
      } catch (e) {
        logger.warn(`Unable to parse JSON from cache at ${cacheFilePath}`);
      }
    }
    if (existingCache.detectionStartTimestamp < await detectionMinimalTimestamp(workspace.rootDirPath)) {
      existingCache = {
        detectionStartTimestamp: 0,
        components: []
      };
    }
    const changedAbsoluteFilePaths = absoluteFilePaths.filter(
      (absoluteFilePath) => {
        const entry = workspace.reader.readSync(absoluteFilePath);
        return entry?.kind === "file" && (options.forceRefresh || entry.lastModifiedMillis() >= existingCache.detectionStartTimestamp);
      }
    );
    const refreshedFilePaths = new Set(
      changedAbsoluteFilePaths.map(
        (absoluteFilePath) => path.relative(workspace.rootDirPath, absoluteFilePath).replace(/\\/g, "/")
      )
    );
    const recycledComponents = existingCache.components.filter(
      ({ componentId }) => {
        const filePath = decodeComponentId(componentId).filePath;
        return filePathsSet.has(filePath) && !refreshedFilePaths.has(filePath);
      }
    );
    const refreshedComponents = await detectComponentsCore(
      logger,
      workspace,
      frameworkPlugin,
      typeAnalyzer,
      changedAbsoluteFilePaths
    );
    const components = [...recycledComponents, ...refreshedComponents];
    if (!options.filePaths) {
      await fs$1.mkdirp(path.dirname(cacheFilePath));
      const updatedCache = {
        detectionStartTimestamp,
        components
      };
      await fs$1.writeFile(cacheFilePath, JSON.stringify(updatedCache));
    }
    return { components };
  });
}
async function detectComponentsCore(logger, workspace, frameworkPlugin, typeAnalyzer, changedAbsoluteFilePaths) {
  const components = [];
  if (changedAbsoluteFilePaths.length === 0) {
    return components;
  }
  logger.debug(
    `Running component detection with file paths:
- ${changedAbsoluteFilePaths.map(
      (absoluteFilePath) => path.relative(workspace.rootDirPath, absoluteFilePath)
    ).join("\n- ")}`
  );
  const found = await frameworkPlugin.detectComponents(
    workspace.reader,
    typeAnalyzer,
    changedAbsoluteFilePaths
  );
  logger.debug(`Done running component detection`);
  for (const component of found) {
    components.push(detectedComponentToApiComponent(component));
  }
  return components;
}
function detectedComponentToApiComponent(component) {
  const [start, end] = component.offsets[0];
  return {
    componentId: component.componentId,
    start,
    end,
    info: component.info.kind === "component" ? {
      kind: "component",
      exported: component.info.exported
    } : {
      kind: "story",
      args: component.info.args,
      associatedComponentId: component.info.associatedComponent.componentId
    }
  };
}
async function detectionMinimalTimestamp(rootDirPath) {
  const nodeModulesPath = path.join(rootDirPath, "node_modules");
  let lastModifiedMillis = 0;
  if (await fs$1.pathExists(nodeModulesPath)) {
    for (const subdirectory of await fs$1.readdir(nodeModulesPath)) {
      const subdirectoryPath = path.join(nodeModulesPath, subdirectory);
      const stat = await fs$1.lstat(subdirectoryPath);
      if (stat.isDirectory() || stat.isSymbolicLink()) {
        lastModifiedMillis = Math.max(lastModifiedMillis, stat.mtimeMs);
      }
    }
  }
  for (const f of FILES_REQUIRING_REDETECTION) {
    const filePath = path.join(rootDirPath, f);
    if (await fs$1.pathExists(filePath)) {
      const stat = await fs$1.stat(filePath);
      lastModifiedMillis = Math.max(lastModifiedMillis, stat.mtimeMs);
    }
  }
  return lastModifiedMillis;
}

var __accessCheck = (obj, member, msg) => {
  if (!member.has(obj))
    throw TypeError("Cannot " + msg);
};
var __privateGet = (obj, member, getter) => {
  __accessCheck(obj, member, "read from private field");
  return getter ? getter.call(obj) : member.get(obj);
};
var __privateAdd = (obj, member, value) => {
  if (member.has(obj))
    throw TypeError("Cannot add the same private member more than once");
  member instanceof WeakSet ? member.add(obj) : member.set(obj, value);
};
var __privateSet = (obj, member, value, setter) => {
  __accessCheck(obj, member, "write to private field");
  setter ? setter.call(obj, value) : member.set(obj, value);
  return value;
};
var _server;
class Server {
  constructor(options) {
    this.serverTerminator = null;
    __privateAdd(this, _server, null);
    this.logger = options.logger;
    const app = express();
    app.use((_req, res, next) => {
      res.setHeader("Cache-Control", "max-age=0, must-revalidate");
      next();
    });
    for (const middleware of options.middlewares) {
      app.use(middleware);
    }
    this.app = app;
  }
  get server() {
    return __privateGet(this, _server);
  }
  async start(port) {
    return new Promise((resolve) => {
      const server = this.app.listen(port, () => {
        this.logger.info(
          `Preview.js server running at http://localhost:${port}.`
        );
        resolve(server);
      });
      this.serverTerminator = createHttpTerminator({
        server,
        gracefulTerminationTimeout: 0
      });
      __privateSet(this, _server, server);
    });
  }
  async stop() {
    if (this.serverTerminator) {
      await this.serverTerminator.terminate();
      this.serverTerminator = null;
    }
    __privateSet(this, _server, null);
    this.logger.info(`Preview.js server stopped.`);
  }
}
_server = new WeakMap();

const extensions = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
const importRegexp = /import\s+([\S]+)\s+from\s+('|")([\S]+)\.(css|less|sass|scss|styl|stylus|postcss)(\?[\S]*)?&?('|")/g;
function cssModulesWithoutSuffixPlugin() {
  return {
    name: "previewjs:css-modules-without-suffix",
    async transform(code, id) {
      id = id.split("?")[0];
      if (!extensions.has(path.extname(id))) {
        return;
      }
      return code.replace(
        importRegexp,
        'import $1 from "$3.$4?$5&.module.css"'
      );
    }
  };
}

const jsExtensions$1 = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
function exportToplevelPlugin() {
  return {
    name: "previewjs:export-toplevel",
    enforce: "post",
    transform: async function(code, id) {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      id = id.split("?", 2)[0];
      const extension = path.extname(id);
      if (!jsExtensions$1.has(extension)) {
        return null;
      }
      try {
        const topLevelEntityNames = findTopLevelEntityNames(code);
        return `${code};
          export {
          ${topLevelEntityNames.map((c) => `${c} as __previewjs__${c},`).join("")}
        }`;
      } catch (e) {
        throw new Error(`Unable to parse ${id}: ${e}`);
      }
    }
  };
}
function findTopLevelEntityNames(source) {
  const parsed = Parser.extend(jsx()).parse(source, {
    ecmaVersion: "latest",
    sourceType: "module"
  });
  const topLevelEntityNames = [];
  for (const statement of parsed.body || []) {
    if (statement.type === "VariableDeclaration") {
      for (const declaration of statement.declarations) {
        if (declaration.type === "VariableDeclarator") {
          addIfIdentifier(topLevelEntityNames, declaration.id);
        }
      }
    }
    if (statement.type === "FunctionDeclaration") {
      addIfIdentifier(topLevelEntityNames, statement.id);
    }
    if (statement.type === "ClassDeclaration") {
      addIfIdentifier(topLevelEntityNames, statement.id);
    }
    if (statement.type === "ExportDefaultDeclaration") {
      if (statement.declaration?.id) {
        addIfIdentifier(topLevelEntityNames, statement.declaration.id);
      }
    }
    if (statement.type === "ExportNamedDeclaration") {
      if (statement.declaration?.id) {
        addIfIdentifier(topLevelEntityNames, statement.declaration.id);
      }
      if (statement.declaration?.declarations) {
        for (const declarator of statement.declaration.declarations) {
          addIfIdentifier(topLevelEntityNames, declarator.id);
        }
      }
    }
  }
  return [...new Set(topLevelEntityNames)];
}
function addIfIdentifier(array, id) {
  if (id && id.type === "Identifier") {
    const name = id.name;
    if (name.endsWith("_default")) {
      return;
    }
    array.push(name);
  }
}

function localEval() {
  return {
    name: "previewjs:local-eval",
    enforce: "post",
    transform: (code, id) => {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.indexOf("?") !== -1) {
        return null;
      }
      return code + // Note: This is an uppercase function so that it "looks like" a React component.
      // This needed to prevent breaking React Refresh, which expects all exports to
      // be React components.
      `
export const PreviewJsEvaluateLocally = async (autogenCallbackPropsSource, propsAssignmentSource, __jsxFactory__) => {
  let autogenCallbackProps = {};
  eval(autogenCallbackPropsSource);
  let properties = {};
  eval(propsAssignmentSource);
  return { autogenCallbackProps, properties };
}
`;
    }
  };
}

function publicAssetImportPluginPlugin({
  rootDirPath,
  publicDir
}) {
  return {
    name: "previewjs:public-asset-import",
    resolveId: async (source, importer) => {
      if (!importer || !source.startsWith("/") || source.startsWith("/__previewjs__/")) {
        return;
      }
      const publicDirAbsolutePath = path.join(rootDirPath, publicDir);
      const potentialPublicFilePath = path.join(publicDirAbsolutePath, source);
      if (!potentialPublicFilePath.startsWith(publicDirAbsolutePath + path.sep)) {
        return;
      }
      if (!await fs$1.pathExists(potentialPublicFilePath)) {
        return;
      }
      return source;
    }
  };
}

const VIRTUAL_PREFIX = `/@previewjs-virtual:`;
const jsExtensions = /* @__PURE__ */ new Set([".js", ".jsx", ".ts", ".tsx"]);
function virtualPlugin(options) {
  const { reader, rootDirPath } = options;
  return {
    name: "previewjs:virtual-fs",
    resolveId: async function(id, importer) {
      const virtualImporter = importer?.startsWith(VIRTUAL_PREFIX) || false;
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.startsWith(VIRTUAL_PREFIX)) {
        id = id.slice(VIRTUAL_PREFIX.length);
      }
      id = id.split("?", 2)[0];
      const extension = path.extname(id);
      let absoluteId;
      if (path.isAbsolute(id)) {
        absoluteId = id;
      } else {
        if (!importer) {
          return null;
        }
        if (virtualImporter) {
          importer = importer.slice(VIRTUAL_PREFIX.length);
        }
        if (extension && !jsExtensions.has(extension) && extension !== ".svg") {
          return path.join(path.dirname(importer), id);
        }
        absoluteId = path.join(path.dirname(importer), id);
      }
      const resolved = await resolveAbsoluteModuleId(absoluteId);
      if (resolved) {
        const [absoluteFilePath] = resolved;
        if (await fs$1.pathExists(absoluteFilePath) && !virtualImporter) {
          return null;
        }
        if (!absoluteId.endsWith(extension)) {
          absoluteId += extension;
        }
        const resolvedId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
        return resolvedId;
      }
      return null;
    },
    load: async function(id) {
      if (id.indexOf(`/node_modules/`) !== -1) {
        return null;
      }
      if (id.startsWith(VIRTUAL_PREFIX)) {
        id = id.slice(VIRTUAL_PREFIX.length);
      }
      const resolved = await resolveAbsoluteModuleId(id);
      if (!resolved) {
        return null;
      }
      const [absoluteFilePath, entry] = resolved;
      let isAllowed = false;
      for (const allowedAbsolutePath of options.allowedAbsolutePaths) {
        const relativePath = path.relative(
          allowedAbsolutePath,
          absoluteFilePath
        );
        if (relativePath && !relativePath.startsWith("..") && !path.isAbsolute(relativePath)) {
          isAllowed = true;
          break;
        }
      }
      if (!isAllowed) {
        options.logger.error(
          `Attempted access to ${absoluteFilePath} which is outside of allowed directories. See https://previewjs.com/docs/config and https://vitejs.dev/config/server-options.html#server-fs-allow for more information.`
        );
        return null;
      }
      if (entry.kind !== "file") {
        options.logger.error(`Unable to read file from ${absoluteFilePath}`);
        return null;
      }
      const source = await entry.read();
      const fileExtension = path.extname(absoluteFilePath);
      if (!jsExtensions.has(fileExtension)) {
        return source;
      }
      const moduleExtension = path.extname(id);
      return {
        code: (
          // We run an esbuild transform for .js or no extension
          // because it may include JSX. Otherwise, let plugins
          // decide whether to use esbuild or not.
          !id.includes("__previewjs_internal__") && (moduleExtension === "" || moduleExtension === ".js") ? (await transformWithEsbuild(source, absoluteFilePath, {
            loader: absoluteFilePath.endsWith(".ts") ? "ts" : "tsx",
            format: "esm",
            target: "es2020",
            sourcefile: path.relative(rootDirPath, absoluteFilePath),
            ...options.esbuildOptions
          })).code : source
        )
      };
    },
    transform: (code, id) => {
      if (!id.startsWith(VIRTUAL_PREFIX)) {
        return null;
      }
      return {
        code,
        map: {
          mappings: "",
          names: [],
          sources: [],
          version: 3
        }
      };
    },
    handleHotUpdate: async function(context) {
      const moduleGraph = options.moduleGraph();
      if (!moduleGraph) {
        return;
      }
      const filePath = path.relative(rootDirPath, context.file);
      const absoluteFilePath = path.join(rootDirPath, filePath);
      const entry = await reader.read(absoluteFilePath);
      if (!entry || entry.kind !== "file") {
        return;
      }
      const virtualModuleId = VIRTUAL_PREFIX + absoluteFilePath.replace(/\\/g, "/");
      const node = moduleGraph.getModuleById(virtualModuleId);
      return node && [node];
    }
  };
  async function resolveAbsoluteModuleId(id) {
    return await resolveBaseFilePath(id) || await resolveBaseFilePath(path.join(rootDirPath, id));
  }
  async function resolveBaseFilePath(baseFilePath) {
    for (const suffix of [
      "",
      ...jsExtensions,
      ...[...jsExtensions].map((extension) => `${path.sep}index${extension}`)
    ]) {
      const absoluteFilePath = `${baseFilePath}${suffix}`;
      const entry = await reader.read(absoluteFilePath);
      if (entry?.kind === "file") {
        return [absoluteFilePath, entry];
      }
    }
    return null;
  }
}

class ViteManager {
  constructor(options) {
    this.options = options;
    const router = express.Router();
    router.use(async (req, res, next) => {
      try {
        const viteServer = await this.awaitViteServerReady();
        viteServer.middlewares(req, res, next);
      } catch (e) {
        this.options.logger.error(`Routing error: ${e}`);
        res.status(500).end(`Error: ${e}`);
      }
    });
    this.middleware = router;
  }
  async loadIndexHtml(url, componentId) {
    const template = await fs$1.readFile(
      this.options.shadowHtmlFilePath,
      "utf-8"
    );
    const viteServer = await this.awaitViteServerReady();
    const { filePath } = decodeComponentId(componentId);
    const componentPath = filePath.replace(/\\/g, "/");
    const wrapper = this.options.config.wrapper;
    const wrapperPath = wrapper && await fs$1.pathExists(path.join(this.options.rootDirPath, wrapper.path)) ? wrapper.path.replace(/\\/g, "/") : null;
    return await viteServer.transformIndexHtml(
      url,
      template.replace(/%([^%]+)%/gi, (matched) => {
        switch (matched) {
          case "%INIT_PREVIEW_BLOCK%":
            return `
    let latestComponentModule;
    let latestWrapperModule;
    let refresh;
    
    import.meta.hot.accept(["/${componentPath}"], ([componentModule]) => {
      if (componentModule && refresh) {
        latestComponentModule = componentModule;
        refresh(latestComponentModule, latestWrapperModule);
      }
    });

    ${wrapperPath ? `
    const wrapperModulePromise = import(/* @vite-ignore */ "/${wrapperPath}");
    import.meta.hot.accept(["/${wrapperPath}"], ([wrapperModule]) => {
      if (wrapperModule && refresh) {
        latestWrapperModule = wrapperModule;
        refresh(latestComponentModule, latestWrapperModule);
      }
    });
    ` : `
    const wrapperModulePromise = Promise.all([${this.options.detectedGlobalCssFilePaths.map(
              (cssFilePath) => `import(/* @vite-ignore */ "/${cssFilePath.replace(
                /\\/g,
                "/"
              )}").catch(() => null)`
            ).join(",")}]).then(() => null);
    `}

    // Important: the wrapper must be loaded first as it may monkey-patch
    // modules imported by the component module.
    wrapperModulePromise.then(wrapperModule => {
      latestWrapperModule = wrapperModule;
      import(/* @vite-ignore */ "/${componentPath}").then(componentModule => {
        latestComponentModule = componentModule;
        refresh = initPreview({
          componentModule,
          componentId: ${JSON.stringify(componentId)},
          wrapperModule,
          wrapperName: ${JSON.stringify(wrapper?.componentName || null)},
        });
      });
    });
  `;
          default:
            throw new Error(`Unknown template key: ${matched}`);
        }
      })
    );
  }
  async awaitViteServerReady() {
    await this.viteStartupPromise;
    if (!this.viteServer) {
      throw new Error(`Vite server is not running.`);
    }
    return this.viteServer;
  }
  async start(server, port) {
    let resolveViteStartupPromise;
    this.viteStartupPromise = new Promise((resolve) => {
      resolveViteStartupPromise = resolve;
    });
    const tsInferredAlias = [];
    let tsConfig = null;
    for (const potentialTsConfigFileName of [
      "tsconfig.json",
      "jsconfig.json"
    ]) {
      const potentialTsConfigFilePath = path.join(
        this.options.rootDirPath,
        potentialTsConfigFileName
      );
      if (await fs$1.pathExists(potentialTsConfigFilePath)) {
        tsConfig = loadTsconfig(potentialTsConfigFilePath) || null;
        if (tsConfig) {
          break;
        }
      }
    }
    this.options.logger.debug(
      `Loaded ts/jsconfig: ${JSON.stringify(tsConfig || null, null, 2)}`
    );
    const baseUrl = tsConfig?.compilerOptions?.baseUrl || "";
    const tsConfigPaths = tsConfig?.compilerOptions?.paths || {};
    let baseAlias = baseUrl.startsWith("./") ? baseUrl.substring(1) : baseUrl;
    if (baseAlias && !baseAlias.endsWith("/")) {
      baseAlias += "/";
    }
    for (const [match, mapping] of Object.entries(tsConfigPaths)) {
      const firstMapping = mapping[0];
      if (!firstMapping) {
        continue;
      }
      const matchNoWildcard = match.endsWith("/*") ? match.slice(0, match.length - 2) : match;
      const firstMappingNoWildcard = firstMapping.endsWith("/*") ? firstMapping.slice(0, firstMapping.length - 2) : firstMapping;
      tsInferredAlias.push({
        find: matchNoWildcard,
        replacement: path.join(
          this.options.rootDirPath,
          baseUrl,
          firstMappingNoWildcard
        )
      });
    }
    const existingViteConfig = await vite.loadConfigFromFile(
      {
        command: "serve",
        mode: "development"
      },
      void 0,
      this.options.rootDirPath
    );
    const defaultLogger = vite.createLogger(
      viteLogLevelFromPinoLogger(this.options.logger)
    );
    const frameworkPluginViteConfig = this.options.frameworkPlugin.viteConfig(
      await flattenPlugins([
        ...existingViteConfig?.config.plugins || [],
        ...this.options.config.vite?.plugins || []
      ])
    );
    const publicDir = this.options.config.vite?.publicDir || existingViteConfig?.config.publicDir || frameworkPluginViteConfig.publicDir || this.options.config.publicDir;
    const vitePlugins = [
      viteTsconfigPaths({
        root: this.options.rootDirPath
      }),
      virtualPlugin({
        logger: this.options.logger,
        reader: this.options.reader,
        rootDirPath: this.options.rootDirPath,
        allowedAbsolutePaths: this.options.config.vite?.server?.fs?.allow || [
          searchForWorkspaceRoot(this.options.rootDirPath)
        ],
        moduleGraph: () => this.viteServer?.moduleGraph || null,
        esbuildOptions: frameworkPluginViteConfig.esbuild || {}
      }),
      localEval(),
      exportToplevelPlugin(),
      fakeExportedTypesPlugin({
        readFile: (absoluteFilePath) => this.options.reader.read(absoluteFilePath).then((entry) => {
          if (entry?.kind !== "file") {
            return null;
          }
          return entry.read();
        })
      }),
      cssModulesWithoutSuffixPlugin(),
      publicAssetImportPluginPlugin({
        rootDirPath: this.options.rootDirPath,
        publicDir
      }),
      frameworkPluginViteConfig.plugins
    ];
    const plugins = await Promise.all(
      vitePlugins.flat().map(async (pluginOrPromise) => {
        const plugin = await pluginOrPromise;
        if (!plugin || Array.isArray(plugin) || !plugin.handleHotUpdate) {
          return plugin;
        }
        const handleHotUpdate = typeof plugin.handleHotUpdate === "function" ? plugin.handleHotUpdate : plugin.handleHotUpdate.handler;
        return {
          ...plugin,
          handleHotUpdate: async (ctx) => {
            return handleHotUpdate({
              ...ctx,
              read: async () => {
                const entry = await this.options.reader.read(ctx.file);
                if (entry?.kind !== "file") {
                  return ctx.read();
                }
                return entry.read();
              }
            });
          }
        };
      })
    );
    this.options.logger.debug(`Creating Vite server`);
    const viteServerPromise = vite.createServer({
      ...frameworkPluginViteConfig,
      ...existingViteConfig?.config,
      ...this.options.config.vite,
      configFile: false,
      root: this.options.rootDirPath,
      base: "/preview/",
      optimizeDeps: {
        entries: []
      },
      server: {
        middlewareMode: true,
        hmr: {
          overlay: false,
          server,
          clientPort: port,
          ...typeof this.options.config.vite?.server?.hmr === "object" ? this.options.config.vite?.server?.hmr : {}
        },
        fs: {
          strict: false,
          ...this.options.config.vite?.server?.fs || {}
        },
        ...this.options.config.vite?.server
      },
      customLogger: {
        info: defaultLogger.info,
        warn: defaultLogger.warn,
        error: (msg, options) => {
          if (!msg.startsWith("\x1B[31mInternal server error")) {
            this.viteServer?.ws.send({
              type: "error",
              err: {
                message: msg,
                stack: ""
              }
            });
          }
          defaultLogger.error(msg, options);
        },
        warnOnce: defaultLogger.warnOnce,
        clearScreen: () => {
        },
        hasWarned: defaultLogger.hasWarned,
        hasErrorLogged: defaultLogger.hasErrorLogged
      },
      clearScreen: false,
      cacheDir: this.options.config.vite?.cacheDir || existingViteConfig?.config.cacheDir || this.options.cacheDir,
      publicDir,
      plugins,
      define: {
        __filename: void 0,
        __dirname: void 0,
        ...frameworkPluginViteConfig.define,
        ...existingViteConfig?.config.define,
        ...this.options.config.vite?.define
      },
      resolve: {
        ...existingViteConfig?.config.resolve,
        ...this.options.config.vite?.resolve,
        alias: [
          // First defined rules are applied first, therefore highest priority should come first.
          ...viteAliasToRollupAliasEntries(
            this.options.config.vite?.resolve?.alias
          ),
          ...viteAliasToRollupAliasEntries(this.options.config.alias),
          ...viteAliasToRollupAliasEntries(
            existingViteConfig?.config.resolve?.alias
          ),
          ...tsInferredAlias,
          {
            find: /^~(.*)/,
            replacement: baseAlias + "$1"
          },
          {
            find: "@",
            replacement: baseAlias
          },
          ...viteAliasToRollupAliasEntries(
            frameworkPluginViteConfig.resolve?.alias
          )
        ]
      }
    });
    viteServerPromise.catch((e) => {
      this.options.logger.error(`Vite startup error: ${e}`);
      resolveViteStartupPromise();
      delete this.viteStartupPromise;
    });
    this.viteServer = await viteServerPromise;
    delete this.viteStartupPromise;
    resolveViteStartupPromise();
    this.options.logger.debug(`Done starting Vite server`);
    this.viteServer.watcher.addListener("change", (path2) => {
      this.viteServer?.ws.send({
        type: "custom",
        event: "previewjs-file-changed",
        data: {
          path: path2
        }
      });
    });
  }
  async stop() {
    await this.viteStartupPromise;
    if (this.viteServer) {
      await this.viteServer.close();
      delete this.viteServer;
    }
  }
  getConfig() {
    return this.options.config;
  }
  triggerReload(absoluteFilePath) {
    if (!this.viteServer) {
      return;
    }
    for (const onChange of this.viteServer.watcher.listeners("change")) {
      onChange(absoluteFilePath);
    }
  }
  triggerFullReload() {
    this.viteServer?.moduleGraph.invalidateAll();
    this.viteServer?.ws.send({
      type: "full-reload"
    });
  }
}
function viteAliasToRollupAliasEntries(alias) {
  if (!alias) {
    return [];
  }
  if (Array.isArray(alias)) {
    return alias;
  } else {
    return Object.entries(alias).map(([find, replacement]) => ({
      find,
      replacement
    }));
  }
}
async function flattenPlugins(pluginOptions) {
  const plugins = [];
  for (const pluginOption of await Promise.all(pluginOptions)) {
    if (!pluginOption) {
      continue;
    }
    if (Array.isArray(pluginOption)) {
      plugins.push(...await flattenPlugins(pluginOption));
    } else {
      plugins.push(pluginOption);
    }
  }
  return plugins;
}
function viteLogLevelFromPinoLogger(logger) {
  switch (logger.level) {
    case "fatal":
      return "silent";
    case "error":
      return "error";
    case "warn":
      return "warn";
    case "info":
      return "info";
    case "debug":
      return "info";
    case "trace":
      return "info";
    case "silent":
      return "silent";
    default:
      logger.warn(`Unknown log level: ${logger.level}`);
      return "info";
  }
}

const POSTCSS_CONFIG_FILE = [
  ".postcssrc",
  ".postcssrc.json",
  ".postcssrc.yml",
  ".postcssrc.js",
  ".postcssrc.mjs",
  ".postcssrc.cjs",
  ".postcssrc.ts",
  "postcss.config.js",
  "postcss.config.mjs",
  "postcss.config.cjs",
  "postcss.config.ts"
];
const GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT = [
  "index",
  "global",
  "globals",
  "style",
  "styles",
  "app"
];
const GLOBAL_CSS_EXTS = ["css", "sass", "scss", "less", "styl", "stylus"];
const GLOBAL_CSS_FILE = GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT.flatMap(
  (fileName) => GLOBAL_CSS_EXTS.map((ext) => `${fileName}.${ext}`)
);
const FILES_REQUIRING_RESTART = /* @__PURE__ */ new Set([
  PREVIEW_CONFIG_NAME,
  ...FILES_REQUIRING_REDETECTION,
  ...POSTCSS_CONFIG_FILE,
  ...GLOBAL_CSS_FILE,
  "vite.config.js",
  "vite.config.ts",
  // TODO: Make plugins contribute files requiring restart to make core agnostic of Svelte config files.
  "svelte.config.js"
]);
const SHUTDOWN_CHECK_INTERVAL = 3e3;
const SHUTDOWN_AFTER_INACTIVITY = 1e4;
class Previewer {
  constructor(options) {
    this.options = options;
    this.appServer = null;
    this.viteManager = null;
    this.status = { kind: "stopped" };
    this.shutdownCheckInterval = null;
    this.disposeObserver = null;
    this.config = null;
    this.lastPingTimestamp = 0;
    this.onFileChangeListener = {
      onChange: (absoluteFilePath, info) => {
        absoluteFilePath = path.resolve(absoluteFilePath);
        if (!info.virtual && this.config?.wrapper && absoluteFilePath === path.resolve(this.options.rootDirPath, this.config.wrapper.path)) {
          this.viteManager?.triggerFullReload();
        }
        if (!info.virtual && FILES_REQUIRING_RESTART.has(path.basename(absoluteFilePath))) {
          if (this.status.kind === "starting" || this.status.kind === "started") {
            const port = this.status.port;
            this.options.logger.info(
              "New dependencies were detected. Restarting..."
            );
            this.stop({
              restarting: true
            }).then(async () => {
              await this.start(async () => port, { restarting: true });
            }).catch(this.options.logger.error);
          }
          return;
        }
        if (info.virtual) {
          this.viteManager?.triggerReload(absoluteFilePath);
          this.viteManager?.triggerReload(absoluteFilePath + ".ts");
        } else if (this.options.onFileChanged) {
          this.options.onFileChanged(absoluteFilePath);
        }
      }
    };
    this.transformingReader = createStackedReader([
      options.reader,
      createFileSystemReader({
        mapping: {
          from: options.frameworkPlugin.previewDirPath,
          to: path.join(
            options.rootDirPath,
            "__previewjs_internal__",
            "renderer"
          )
        },
        watch: false
      }),
      createFileSystemReader({
        mapping: {
          from: options.previewDirPath,
          to: options.rootDirPath
        },
        watch: false
      })
    ]);
  }
  async start(allocatePort, options = {}) {
    let port;
    const statusBeforeStart = this.status;
    switch (statusBeforeStart.kind) {
      case "starting":
        port = statusBeforeStart.port;
        try {
          await statusBeforeStart.promise;
        } catch (e) {
          this.options.logger.error(e);
          this.status = {
            kind: "stopped"
          };
          await this.startFromStopped(port);
        }
        break;
      case "started":
        if (this.shutdownCheckInterval) {
          clearInterval(this.shutdownCheckInterval);
          this.shutdownCheckInterval = null;
        }
        port = statusBeforeStart.port;
        break;
      case "stopping":
        port = statusBeforeStart.port;
        try {
          await statusBeforeStart.promise;
        } catch (e) {
          this.options.logger.error(e);
          this.status = {
            kind: "stopped"
          };
        }
        await this.start(async () => port, options);
        break;
      case "stopped":
        port = await allocatePort();
        await this.startFromStopped(port, options);
        break;
      default:
        throw assertNever(statusBeforeStart);
    }
    return port;
  }
  async startFromStopped(port, { restarting } = {}) {
    this.status = {
      kind: "starting",
      port,
      promise: (async () => {
        process.chdir(this.options.rootDirPath);
        const configFromProject = await readConfig(this.options.rootDirPath);
        const config = this.config = {
          ...configFromProject,
          wrapper: configFromProject.wrapper || {
            path: this.options.frameworkPlugin.defaultWrapperPath
          }
        };
        const globalCssAbsoluteFilePaths = await findFiles(
          this.options.rootDirPath,
          `**/@(${GLOBAL_CSS_FILE_NAMES_WITHOUT_EXT.join(
            "|"
          )}).@(${GLOBAL_CSS_EXTS.join("|")})`,
          {
            maxDepth: 3
          }
        );
        const router = express.Router();
        router.get(/^\/preview\/.*:[^/]+\/$/, async (req, res) => {
          const componentId = req.path.substring(9, req.path.length - 1);
          if (req.header("Accept") === "text/x-vite-ping") {
            res.writeHead(204).end();
            return;
          }
          if (!this.viteManager) {
            res.status(404).end(`Uh-Oh! Vite server is not running.`);
            return;
          }
          try {
            res.status(200).set({ "Content-Type": "text/html" }).end(
              await this.viteManager.loadIndexHtml(
                req.originalUrl,
                componentId
              )
            );
          } catch (e) {
            res.status(500).set({ "Content-Type": "text/html" }).end(
              `<html>
              <head>
                <style>
                  body {
                    background: #FCA5A5
                  }
                  pre {
                    font-family: source-code-pro, Menlo, Monaco, Consolas, 'Courier New',
                    monospace;
                    font-size: 12px;
                    line-height: 1.5em;
                    color: #7F1D1D;
                  }
                </style>
              </head>
              <body>
                <pre>${escape(`${e}` || "An unknown error has occurred")}</pre>
              </body>
            </html>`
            );
          }
        });
        router.use("/ping", async (req, res) => {
          this.lastPingTimestamp = Date.now();
          res.json(
            JSON.stringify({
              pong: "match!"
            })
          );
        });
        this.appServer = new Server({
          logger: this.options.logger,
          middlewares: [
            ...this.options.middlewares || [],
            router,
            (req, res, next) => {
              this.viteManager?.middleware(req, res, next);
            }
          ]
        });
        if (!restarting) {
          if (this.transformingReader.observe) {
            this.disposeObserver = await this.transformingReader.observe(
              this.options.rootDirPath
            );
          }
          this.transformingReader.listeners.add(this.onFileChangeListener);
        }
        this.viteManager = new ViteManager({
          rootDirPath: this.options.rootDirPath,
          shadowHtmlFilePath: path.join(
            this.options.previewDirPath,
            "index.html"
          ),
          detectedGlobalCssFilePaths: globalCssAbsoluteFilePaths.map(
            (absoluteFilePath) => path.relative(this.options.rootDirPath, absoluteFilePath)
          ),
          reader: this.transformingReader,
          cacheDir: path.join(getCacheDir(this.options.rootDirPath), "vite"),
          config,
          logger: this.options.logger,
          frameworkPlugin: this.options.frameworkPlugin
        });
        this.options.logger.debug(`Starting server`);
        const server = await this.appServer.start(port);
        this.options.logger.debug(`Starting Vite manager`);
        this.viteManager.start(server, port).catch((e) => {
          this.options.logger.error(`Vite manager failed to start: ${e}`);
          this.stop();
        });
        this.options.logger.debug(`Previewer ready`);
        this.status = {
          kind: "started",
          port
        };
      })()
    };
    await this.status.promise;
    for (let i = 0; ; i++) {
      try {
        await axios.get(`http://localhost:${port}`);
        break;
      } catch (e) {
        if (i === 10) {
          throw e;
        } else {
          await new Promise((resolve) => setTimeout(resolve, 50));
        }
      }
    }
  }
  async stop(options = {}) {
    if (this.status.kind === "starting") {
      try {
        await this.status.promise;
      } catch {
        this.status = {
          kind: "stopped"
        };
      }
    }
    if (this.status.kind !== "started") {
      return;
    }
    if (options.onceUnused) {
      if (!await this.shutdownCheck()) {
        this.shutdownCheckInterval = setInterval(() => {
          this.shutdownCheck().catch(this.options.logger.error);
        }, SHUTDOWN_CHECK_INTERVAL);
      }
    } else {
      await this.stopNow(options);
    }
  }
  async shutdownCheck() {
    if (this.lastPingTimestamp + SHUTDOWN_AFTER_INACTIVITY > Date.now()) {
      return false;
    }
    if (this.shutdownCheckInterval) {
      clearInterval(this.shutdownCheckInterval);
      this.shutdownCheckInterval = null;
    }
    await this.stopNow();
    return true;
  }
  async stopNow({
    restarting
  } = {}) {
    if (this.status.kind === "starting") {
      await this.status.promise;
    }
    if (this.status.kind !== "started") {
      return;
    }
    this.status = {
      kind: "stopping",
      port: this.status.port,
      promise: (async () => {
        if (!restarting) {
          this.transformingReader.listeners.remove(this.onFileChangeListener);
          if (this.disposeObserver) {
            await this.disposeObserver();
            this.disposeObserver = null;
          }
        }
        await this.viteManager?.stop();
        this.viteManager = null;
        await this.appServer?.stop();
        this.appServer = null;
        this.status = {
          kind: "stopped"
        };
      })()
    };
    await this.status.promise;
  }
}

class ApiRouter {
  constructor(logger) {
    this.logger = logger;
    this.handlers = /* @__PURE__ */ new Map();
    this.registerRPC = (endpoint, handler) => {
      if (this.handlers.has(endpoint.path)) {
        throw new Error(
          `Multiple handlers registered for endpoint '${endpoint.path}'`
        );
      }
      this.handlers.set(endpoint.path, handler);
    };
  }
  async handle(path, request) {
    const handler = this.handlers.get(path);
    if (!handler) {
      return {
        kind: "error",
        message: `No handler is registered for path '${path}'`
      };
    }
    this.logger.debug(`RPC ${path} invoked`);
    try {
      const response = await handler(request);
      return {
        kind: "success",
        response
      };
    } catch (e) {
      this.logger.error(`RPC ${path} failed`, e);
      return {
        kind: "error",
        message: e.message
      };
    }
  }
}

const require$1 = createRequire(import.meta.url);
async function setupFrameworkPlugin({
  rootDirPath,
  frameworkPlugins,
  logger
}) {
  const dependencies = await extractPackageDependencies(logger, rootDirPath);
  for (const candidate of frameworkPlugins) {
    if (await candidate.isCompatible(dependencies)) {
      return candidate.create({
        rootDirPath,
        dependencies,
        logger
      });
    }
  }
  return null;
}
async function extractPackageDependencies(logger, rootDirPath) {
  const packageJsonPath = path.join(rootDirPath, "package.json");
  if (!await fs$1.pathExists(packageJsonPath)) {
    return {};
  }
  const { dependencies, devDependencies, peerDependencies } = JSON.parse(
    await fs$1.readFile(packageJsonPath, "utf8")
  );
  const allDependencies = {
    ...dependencies,
    ...devDependencies,
    ...peerDependencies
  };
  return Object.fromEntries(
    Object.entries(allDependencies).map(
      ([name, version]) => {
        let majorVersion;
        if (typeof version !== "string") {
          majorVersion = 0;
        } else if (version.startsWith("^") || version.startsWith("~")) {
          majorVersion = parseInt(version.slice(1));
        } else {
          majorVersion = parseInt(version);
        }
        const readInstalledVersion = async () => {
          try {
            const moduleEntryPath = findModuleEntryPath(name, rootDirPath);
            let packagePath = moduleEntryPath;
            let packageJsonPath2 = null;
            while (packagePath !== path.dirname(packagePath)) {
              const candidatePackageJsonPath = path.join(
                packagePath,
                "package.json"
              );
              if (fs$1.existsSync(candidatePackageJsonPath)) {
                packageJsonPath2 = candidatePackageJsonPath;
                break;
              }
              packagePath = path.dirname(packagePath);
            }
            if (!packageJsonPath2) {
              throw new Error(
                `No package.json path found from: ${moduleEntryPath}`
              );
            }
            const packageInfo = JSON.parse(
              await fs$1.readFile(packageJsonPath2, "utf8")
            );
            const version2 = packageInfo["version"];
            if (!version2 || typeof version2 !== "string") {
              throw new Error(
                `Invalid version found for package: ${packageJsonPath2}`
              );
            }
            return version2;
          } catch (e) {
            logger.error(
              `Unable to read installed version of package: ${name}`,
              e
            );
            return null;
          }
        };
        return [name, { majorVersion, readInstalledVersion }];
      }
    )
  );
}
function findModuleEntryPath(name, rootDirPath) {
  try {
    return require$1.resolve(name, {
      paths: [rootDirPath]
    });
  } catch (e) {
    if (e.code === "ERR_PACKAGE_PATH_NOT_EXPORTED") {
      const potentialMatch = e.message.match(
        /No "exports" main defined in (.*)(\/|\\)package\.json/
      );
      if (potentialMatch) {
        return potentialMatch[1];
      }
    }
    throw e;
  }
}

const require = createRequire(import.meta.url);
process.on("uncaughtException", (e) => {
  console.error("Encountered an uncaught exception", e);
});
process.on("unhandledRejection", (e) => {
  console.error("Encountered an unhandled promise", e);
});
async function createWorkspace({
  rootDirPath,
  reader,
  frameworkPlugin,
  logger,
  setupEnvironment
}) {
  logger.debug(
    `Creating workspace with framework plugin ${frameworkPlugin.name} from root: ${rootDirPath}`
  );
  const expectedPluginApiVersion = 3;
  if (!frameworkPlugin.pluginApiVersion || frameworkPlugin.pluginApiVersion < expectedPluginApiVersion) {
    throw new Error(
      `Preview.js framework plugin ${frameworkPlugin.name} is incompatible with this version of Preview.js. Please upgrade it.`
    );
  } else if (frameworkPlugin.pluginApiVersion > expectedPluginApiVersion) {
    throw new Error(
      `Preview.js framework plugin ${frameworkPlugin.name} is too recent. Please upgrade Preview.js or use an older version of ${frameworkPlugin.name}.`
    );
  }
  if (frameworkPlugin.transformReader) {
    reader = frameworkPlugin.transformReader(reader);
  }
  const collected = {};
  const typeAnalyzer = createTypeAnalyzer({
    reader,
    rootDirPath,
    collected,
    specialTypes: frameworkPlugin.specialTypes,
    tsCompilerOptions: frameworkPlugin.tsCompilerOptions,
    warn: logger.warn.bind(logger)
  });
  const router = new ApiRouter(logger);
  router.registerRPC(RPCs.ComputeProps, async ({ componentIds }) => {
    logger.debug(`Computing props for components: ${componentIds.join(", ")}`);
    let analyze;
    const detectedComponents = await frameworkPlugin.detectComponents(
      reader,
      typeAnalyzer,
      [
        ...new Set(
          componentIds.map(
            (c) => path.join(rootDirPath, decodeComponentId(c).filePath)
          )
        )
      ]
    );
    logger.debug(`Detected ${detectedComponents.length} components`);
    const componentIdToDetectedComponent = Object.fromEntries(
      detectedComponents.map((c) => [c.componentId, c])
    );
    const components = {};
    let types = {};
    for (const componentId of componentIds) {
      const component = componentIdToDetectedComponent[componentId];
      if (!component) {
        const { filePath, name } = decodeComponentId(componentId);
        throw new Error(`Component ${name} not detected in ${filePath}.`);
      }
      if (component.info.kind === "component") {
        analyze = component.info.analyze;
      } else {
        analyze = component.info.associatedComponent.analyze;
      }
      logger.debug(`Analyzing ${component.info.kind}: ${componentId}`);
      const { propsType: props, types: componentTypes } = await analyze();
      logger.debug(`Done analyzing: ${componentId}`);
      components[componentId] = {
        info: component.info.kind === "component" ? {
          kind: "component",
          exported: component.info.exported
        } : {
          kind: "story",
          args: component.info.args,
          associatedComponentId: component.info.associatedComponent.componentId
        },
        props
      };
      types = { ...types, ...componentTypes };
    }
    return {
      components,
      types
    };
  });
  router.registerRPC(
    RPCs.DetectComponents,
    (options) => detectComponents(logger, workspace, frameworkPlugin, typeAnalyzer, options)
  );
  const middlewares = [
    express.json(),
    async (req, res, next) => {
      if (req.path.startsWith("/api/")) {
        res.json(await router.handle(req.path.substr(5), req.body));
      } else {
        next();
      }
    }
  ];
  const previewer = new Previewer({
    reader,
    rootDirPath,
    // TODO: Use a cleaner approach.
    previewDirPath: path.join(
      path.dirname(path.dirname(require.resolve("@previewjs/iframe"))),
      "preview"
    ),
    frameworkPlugin,
    logger,
    middlewares,
    onFileChanged: (absoluteFilePath) => {
      const filePath = path.relative(rootDirPath, absoluteFilePath);
      for (const name of Object.keys(collected)) {
        if (name.startsWith(`${filePath}:`)) {
          delete collected[name];
        }
      }
    }
  });
  async function localRpc(endpoint, request) {
    const result = await router.handle(endpoint.path, request);
    if (result.kind === "success") {
      return result.response;
    }
    throw new Error(result.message);
  }
  const workspace = {
    rootDirPath,
    reader,
    typeAnalyzer,
    detectComponents: (options = {}) => localRpc(RPCs.DetectComponents, options),
    computeProps: (options) => localRpc(RPCs.ComputeProps, options),
    preview: {
      start: async (allocatePort) => {
        const port = await previewer.start(async () => {
          const port2 = allocatePort ? await allocatePort() : 0;
          return port2 || await getPort({
            port: portNumbers(3140, 4e3)
          });
        });
        return {
          url: () => `http://localhost:${port}`,
          stop: async (options) => {
            await previewer.stop(options);
          }
        };
      }
    },
    dispose: async () => {
      typeAnalyzer.dispose();
    }
  };
  if (setupEnvironment) {
    const environment = await setupEnvironment({
      registerRPC: (endpoint, handler) => router.registerRPC(endpoint, handler),
      workspace
    });
    if (environment.middlewares) {
      middlewares.push(...environment.middlewares);
    }
  }
  return workspace;
}
function findWorkspaceRoot(absoluteFilePath) {
  let dirPath = path.resolve(absoluteFilePath);
  while (dirPath !== path.dirname(dirPath)) {
    if (fs$1.existsSync(path.join(dirPath, "package.json"))) {
      return dirPath;
    }
    dirPath = path.dirname(dirPath);
  }
  return null;
}

export { createWorkspace, findWorkspaceRoot, setupFrameworkPlugin };
